'''
Command line interface.
'''
import argparse
import signal
import contextlib
from . import operation

try:
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
except AttributeError:
    pass

SUPPORTED_FORMATS = [i[1:] for i in operation.supported_suffix()]

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--filter", default={},
                        help="The mongo like query filter over .bson or .json file.")
    parser.add_argument("-o", "--output", default="",
                        help="The output .bson or .json file, default to STDOUT if ignored.")
    parser.add_argument("--output-format", default="", choices=SUPPORTED_FORMATS,
                        help="The output format.")
    parser.add_argument("--input-format", default="", choices=SUPPORTED_FORMATS,
                        help="The input format.")
    parser.add_argument("-n", default=None, type=int,
                        help="The number of entries to output.")
    parser.add_argument("--json-indent",
                        default=None,
                        type=lambda x: int(x) if x.isdigit() else x,
                        help="Whether to indent the output json/jsonl file.")
    parser.add_argument("--json-ensure-ascii",
                        action=argparse.BooleanOptionalAction, default=True,
                        help="Whether to enable the `ensure_ascii` option for output json/jsonl file.")
    parser.add_argument("--json-skipkeys", action=argparse.BooleanOptionalAction, default=False,
                        help="Whether to enable the `skipkeys` option for output json/jsonl file.")
    parser.add_argument("--json-allow-nan", action=argparse.BooleanOptionalAction, default=True,
                        help="Whether to enable the `allow_nan` option for output json/jsonl file.")
    parser.add_argument("--json-sort-keys", action=argparse.BooleanOptionalAction, default=False,
                        help="Whether to enable the `sort_keys` option for output json/jsonl file.")
    parser.add_argument("options", default=[], nargs="*")
    return parser.parse_args()


def main():
    args = vars(parse_args())
    options = args['options']
    output = args['output']
    if not output:
        output_format = args['output_format'] or 'json'
    else:
        output_format = args['output_format'] or operation.get_format(output)
    json_options = {k[5:]: v for k, v in args.items() if k.startswith('json_') and v is not None}
    n = args['n']
    class SafeDict(dict):
        def __missing__(self, key):
            return "unknown"
    out_fd_dict = {}
    empty = True
    with contextlib.ExitStack() as stack:
        for idx, entry in enumerate(operation.query_son(options[0],
                                                        file_format=args['input_format'],
                                                        filters=args['filter'])):
            if n is not None and idx >= n:
                break
            empty = False
            formatted_output = output.format_map(SafeDict(entry)) if output else None
            if formatted_output not in out_fd_dict:
                out_fd_dict[formatted_output] = stack.enter_context(operation.get_output_fileobj(formatted_output, output_format))
            out_fd_dict[formatted_output].write(operation.as_output_format(entry, output_format, json_options=json_options))
        if empty and not ('{' in output and '}' in output):
            stack.enter_context(operation.get_output_fileobj(output, output_format))


if __name__ == '__main__':
    main()



from sonq.query import get_separated_attr, match

data = [
    {"num": 0,
     "name": "Wang",
     "info": {"attack": 100},
     "logs": [{"type": "login"}],
     "tags": ["sun", "moon"]}
]


def test_get_separated_attr():
    assert get_separated_attr(data[0], "name") == (True, "Wang")
    assert get_separated_attr(data[0], "info.attack") == (True, 100)
    assert get_separated_attr(data[0], "info.not_existed") == (False, None)
    assert get_separated_attr(data[0], "logs.0.not_existed") == (False, None)
    assert get_separated_attr(data[0], "logs.0.type") == (True, "login")
    assert get_separated_attr(data[0], "logs.1.type") == (False, None)


def test_simple_int_equal():
    assert match(data[0], {'num': 0})


def test_simple_int_inequal():
    assert not match(data[0], {'num': 1})


def test_simple_nested_equal():
    assert match(data[0], {'info': {'attack': 100}})


def test_simple_nested_inequal():
    assert not match(data[0], {'info': {'attack': 50}})


def test_simple_separated_equal():
    assert match(data[0], {'info.attack': 100})


def test_simple_separated_inequal():
    assert not match(data[0], {'info.attack': 50})


def test_op_eq():
    assert match(data[0], {'info.attack': {'$eq': 100}})
    assert match(data[0], {'tags': {'$eq': "sun"}})
    assert match(data[0], {'tags': {'$eq': ["sun", "moon"]}})


def test_op_ne():
    assert match(data[0], {'info.attack': {'$ne': 101}})
    assert match(data[0], {'tags': {'$ne': "fogs"}})


def test_op_in():
    assert match(data[0], {'name': {'$in': ['Wang', 'Zhao']}})
    assert match(data[0], {'tags': {'$in': ['sun']}})
    assert match(data[0], {'tags': {'$in': [['sun', 'moon']]}})


def test_op_nin():
    assert match(data[0], {'name': {'$nin': ['Li', 'Qian']}})
    assert match(data[0], {'tags': {'$nin': ['saturn']}})
    assert match(data[0], {'tags': {'$nin': [['saturn', 'ceras']]}})


def test_op_compare():
    assert match(data[0], {'info.attack': {'$gt': 99}})
    assert match(data[0], {'info.attack': {'$gte': 100}})
    assert match(data[0], {'info.attack': {'$lt': 101}})
    assert match(data[0], {'info.attack': {'$lte': 100}})


def test_op_and():
    assert match(data[0], {'$and': [
        {'name': 'Wang'},
        {'num': 0},
        {'info.attack': {'$ne': 102}},
    ]})


def test_op_or():
    assert match(data[0], {'$or': [
        {'name': 'Li'},
        {'num': 2},
        {'info.attack': {'$ne': 102}},
    ]})


def test_op_nor():
    assert match(data[0], {'$nor': [
        {'name': 'Li'},
        {'num': 2},
        {'info.attack': {'$eq': 102}},
    ]})


def test_op_not():
    assert match(data[0], {'$not': 
                           {"name": "Li"}
    })


def test_exists():
    assert match(data[0], {'num': {'$exists': True}})
    assert match(data[0], {'non': {'$exists': False}})



import gzip
import sys
import json
import os
import functools
import bson
import bson.json_util
import contextlib
from .query import query

SUPPORTED_FORMATS = ('.bson', '.json', '.jsonl')
SUPPORTED_COMPRESS_EXTS = ('.gz', '.bz2')


def supported_suffix():
    '''
    Get the supported file suffixes.
    '''
    for format in SUPPORTED_FORMATS:
        for compress_ext in ['', *SUPPORTED_COMPRESS_EXTS]:
            yield format + compress_ext


def get_format(filename, format=None):
    '''
    Get the file format from filename or provided format.
    '''
    suported = list(supported_suffix())
    if format:
        if format not in suported and ('.%s' % format) not in suported:
            raise ValueError('Unsupported format: %s' % format)
        return format
    else:
        lower_filename = filename.lower()
        for suffix in suported:
            if lower_filename.endswith(suffix):
                format = suffix[1:]
                break
        else:
            if filename == '-':
                format = 'json'
            else:
                format = 'bson'
    return format


def get_output_fileobj(output, output_format):
    '''
    Get output file object based on output filename and output_format.
    '''
    file_format = get_format(output, output_format)
    open_func = open
    if file_format.endswith('.gz'):
        import gzip
        open_func = gzip.open
    elif file_format.endswith('.bz2'):
        import bz2
        open_func = bz2.open
    if output:
        if file_format.startswith('bson'):
            fd = open_func(output, 'wb')
        else:
            fd = open_func(output, 'wt', encoding='utf-8', newline='\n')
    elif output_format.startswith('bson'):
        fd = sys.stdout.buffer
    else:
        fd = sys.stdout
    return fd


def as_output_format(dict_obj, output_format, json_options=None):
    '''
    Convert dict object to str or byte format w.r.t. output_format.
    '''
    if output_format.startswith('bson'):
        return bson.BSON.encode(dict_obj)
    else:
        if json_options is None:
            json_options = {}
        return '%s\n' % (bson.json_util.dumps(dict_obj, **json_options))


def decode_json_file_iter(fd, *args, **kw):
    '''
    Decode newline splitted json file.
    '''
    for line in fd.readlines():
        if not line:
            continue
        yield bson.json_util.loads(line, *args, **kw)


def query_son(filename, file_format=None, filters=None):
    '''
    Query a file with mongo like query filters.
    '''
    if isinstance(filters, str):
        filters = json.loads(filters)
    filters = filters or {}
    file_format = get_format(filename, format=file_format)
    file_mode = 'rb' if file_format.startswith('bson') else 'rt'
    open_func = open
    if file_format.endswith('.gz'):
        import gzip
        open_func = gzip.open
    elif file_format.endswith('.bz2'):
        import bz2
        open_func = bz2.open
    if filename == '-':
        get_fd = lambda: contextlib.nullcontext(sys.stdin)
    else:
        get_fd = lambda: open_func(filename, file_mode)
    decode_iter = bson.decode_file_iter if file_format.startswith('bson') else decode_json_file_iter
    with get_fd() as fd:
        for obj in query(decode_iter(fd), filters):
            yield obj
